#!/bin/sh

# Regression testing script for MicroC
# Step through a list of files
#  Compile, run, and check the output of each expected-to-work test
#  Compile and check the error of each expected-to-fail test

# Path to the LLVM interpreter
LLI="lli"
#LLI="/usr/local/opt/llvm/bin/lli"

# Path to the microc compiler.  Usually "./microc.native"
# Try "_build/microc.native" if ocamlbuild was unable to create a symbolic link.
MICROC="./lava.native"
#MICROC="_build/microc.native"

# Set time limit for all operations
ulimit -t 30

# Global log
globallog=testall.log
rm -f $globallog

# Error flags
error=0
globalerror=0

# Default keep temp files
keep=1

# Report purpose
pos_pass=0
pos_fail=0
neg_pass=0
neg_fail=0
RED='\033[0;31m'
GREEN='\033[0;32m'
CYAN='\033[0;36m'
WHITE='\033[0m'

Usage() {
    echo "Usage: testall.sh [options] [.mc files]"
    echo "-k    Keep intermediate files"
    echo "-h    Print this help"
    exit 1
}

SignalError() {
    if [ $error -eq 0 ] ; then
	echo "${RED}FAILED${WHITE}"
	error=1
    fi
    echo "  $1"
}

# Compare <outfile> <reffile> <difffile>
# Compares the outfile with reffile.  Differences, if any, written to difffile
Compare() {
    generatedfiles="$generatedfiles $3"
    echo diff -b $1 $2 ">" $3 1>&2
    diff -b "$1" "$2" > "$3" 2>&1 || {
	SignalError "$1 differs from $2"
	echo "FAILED $1 differs from $2" 1>&2
    }
}

CompareFail() {
    generatedfiles="$generatedfiles $3"
    s1=`cat $1`
    # echo "Generated err msg: $s1"
    s1=`cat $2` # This doesn't work, dunno why'
    # echo "Expected err msg: $s2"
    echo diff -b $1 $2 ">" $3 1>&2

    expected="Fatal error:"
    if [ -z "${s1##*$expected*}" ]; then
    # if [[ $s1 =~ .*Fatal error:.* ]];then
    # if [[ $s1 == *"Fatal error: exception"* ]];then 
        echo "${GREEN}Failure as expected${WHITE}"
    else 
        SignalError "$1 doesn't contain an error!"
        echo "Output: $s1" 1>&2
    fi 
}


# Run <args>
# Report the command, run it, and report any errors
Run() {
    echo $* 1>&2
    eval $* || {
	SignalError "${RED}$1 failed on $*${WHITE}"
	return 1
    }
}

# RunFail <args>
# Report the command, run it, and expect an error
RunFail() {
    echo $* 1>&2
    eval $* && {
	SignalError "${RED}failed: $* did not report an error${WHITE}"
	return 1
    }
    return 0
}

Check() {
    error=0
    basename=`echo $1 | sed 's/.*\\///
                             s/.mc//'`
    reffile=`echo $1 | sed 's/.mc$//'`
    basedir="`echo $1 | sed 's/\/[^\/]*$//'`/."

    echo -n "$basename..."
    
    echo 1>&2
    echo "###### Testing $basename" 1>&2

    # Generate target dir
    dirname=$2/testrun/

    generatedfiles=""

    generatedfiles="$generatedfiles ${dirname}${basename}.ll ${dirname}${basename}.out" &&
    Run "$MICROC" "<" $1 ">" "${dirname}${basename}.ll" &&
    Run "$LLI" "${dirname}${basename}.ll" ">" "${dirname}${basename}.out" &&
    Compare ${dirname}${basename}.out ${reffile}.out ${dirname}${basename}.diff

    if [ $error -eq 0 ] ; then
	if [ $keep -eq 0 ] ; then
	    rm -f $generatedfiles
	fi
        echo "${GREEN}OK${WHITE}"
        echo "###### SUCCESS" 1>&2
        ((pos_pass++))
        # echo "Now passed pos tests: $pos_pass"
    else
        # echo -n "${RED}$basename...${WHITE}"
        echo "###### FAILED" 1>&2
        globalerror=$error
        ((pos_fail++))
    fi
}

CheckFail() {
    error=0
    basename=`echo $1 | sed 's/.*\\///
                             s/.mc//'`
    reffile=`echo $1 | sed 's/.mc$//'`
    basedir="`echo $1 | sed 's/\/[^\/]*$//'`/."

    echo -n "$basename..."

    echo 1>&2
    echo "###### Testing $basename" 1>&2

    # Generate target dir
    dirname=$2/testrun/
    locallog=$2/testall.log

    generatedfiles=""

    generatedfiles="$generatedfiles ${dirname}${basename}.err ${dirname}${basename}.diff" &&
    RunFail "$MICROC" "<" $1 "2>" "${dirname}${basename}.err" ">>" $locallog &&
    CompareFail ${dirname}${basename}.err ${reffile}.err ${dirname}${basename}.diff

    # Report the status and clean up the generated files

    if [ $error -eq 0 ] ; then
	if [ $keep -eq 0 ] ; then
	    rm -f $generatedfiles
	fi
        echo "${GREEN}OK${WHITE}"
        echo "###### SUCCESS" 1>&2
        let neg_pass=neg_pass+1
    else
	    # echo -n "${RED}$basename...${WHITE}"
        echo "${RED}###### FAILED${WHITE}" 1>&2
	    globalerror=$error
        let neg_fail=neg_fail+1
    fi
}

while getopts kdpsh c; do
    case $c in
	k) # Keep intermediate files
	    keep=1
	    ;;
	h) # Help
	    Usage
	    ;;
    esac
done

shift `expr $OPTIND - 1`

LLIFail() {
  echo "Could not find the LLVM interpreter \"$LLI\"."
  echo "Check your LLVM installation and/or modify the LLI variable in testall.sh"
  exit 1
}

which "$LLI" >> $globallog || LLIFail


if [ $# -ge 1 ]
then
    folders=$@
else
    # files="tests/**/test-*.mc tests/**/fail-*.mc"
    folders="tests/*"
fi

# Generate temp folders for holding temp files
echo "PWD: $PWD"
for folder in $folders
do 
    temp_folder="$PWD/$folder/testrun/"
    mkdir -p $temp_folder
    rc=$? # return code of mkdir
    if [ $rc != 0 ] 
    then echo "mkdir in $temp_folder exited with code $rc"
    fi
done

# Test folder by folder
for folder in $folders
do 
    echo "\nTesting in $folder"
    # log path
    locallog="$folder/testall.log"
    rm -f $locallog

    # temp file path
    tempfilefolder=$folder/testrun
    echo "Put temp files in $tempfilefolder"
    rm -Rf tempfilefolder
    mkdir -p tempfilefolder

    # test files
    files="$folder/test-*.mc $folder/fail-*.mc"
    echo "Files to work on: $files"

    for file in $files
    do
        case $file in
        *test-*)
            Check $file $folder 2>> $locallog
            ;;
        *fail-*)
            CheckFail $file $folder 2>> $locallog
            ;;
        *)
            echo "${RED}unknown file type $file${WHITE}"
            globalerror=1
            ;;
        esac
    done
done

# reporting
# echo "Passed pos tests: $pos_pass"
# echo "Failed pos tests: $pos_fail"
# echo "Passed neg tests: $neg_pass"
# echo "Failed neg tests: $neg_fail"


exit $globalerror
